#######################################################################################################################
#   
#   Use this script to create the needed Sidra Core metadata for ingesting CSV files into Databricks.
#   Parameters may be fed via a .PSD1 file; then the name of the .PSD1 is fed as parameter to this script. Example:
#       C:\>    .\Define-CSV-Entity.ps1 -ParamsFile "E:\Temp\Define-CSV-Entity-BikesStoreBikes.psd1"
#   Look into the accompanying 'Define-CSV-Entity.psd1' to see what kind of structure the params file should have.
#
#######################################################################################################################



[CmdletBinding()]
param(
    [Parameter()] [string] $ParamsFile = $null,

    [Parameter()] [string] $ApiUrl = $null,
    [Parameter()] [string] $AuthenticationUrl = $null,
    [Parameter()] [string] $OAuthClientID = $null,
    [Parameter()] [string] $OAuthSecret = $null,
    
    [Parameter()] [string] $ProviderName = $null,
    [Parameter()] [string] $ProviderOwner = $null,
    [Parameter()] [string] $ProviderDescription = $null,

    [Parameter()] [string] $EntityName = $null,
    [Parameter()] [string] $EntityRowDelimiter = $null,           #   $null works fine as a default: rows delimited by [Carriage-Return]
    [Parameter()] [string] $EntityFieldDelimiter = $null,         #   \u002C=','    \u003B=';'    \u0009='\t'(TAB)
    [Parameter()] [string] $EntityRegularExpression = $null,      #   "^Bike-Store.Bikes_((?<year>\d{4})-(?<month>\d{2})-(?<day>\d{2}))_((?<hour>\d{2})-(?<minute>\d{2})-(?<second>\d{2})).csv"
    [Parameter()] [string] $EntityFormat = "CSV",
    [Parameter()] [string] $EntityEncoding = "UTF-8",
    [Parameter()] [int]    $EntityHeaderLinesCount = 1,
    [Parameter()] [string] $EntityStartValidDate = "2020-12-23T13:17:02.881Z",
    [Parameter()] [string] $EntityDescription = $null,

    [Parameter()] [PSCustomObject] $AttributesCollection = $null
);

$PsStyle.Formatting.Verbose = $PsStyle.Foreground.FromRgb(0x333333);

#----------------------------------------------------------------------------------------------------------------------
#region     VALIDATE PARAMETERS; TENTATIVELY READ .PSD1 FILE FOR REQUIRED PARAMS THAT ARE MISSING IN COMMAND-LINE
#----------------------------------------------------------------------------------------------------------------------

function ParameterIsMissing {
    param ( $TheParameter );
    if ($null -eq $TheParameter) { return $true; }
    if (-not ($TheParameter)) { return $true; }
    if ($TheParameter.ToString().Length -eq 0) { return $true; }
    if ([string]::IsNullOrWhiteSpace($TheParameter.ToString())) { return $true; }
    if ($TheParameter.ToString().StartsWith("POPULATE-THIS", "CurrentCultureIgnoreCase")) { return $true; }
    return $false;
}
function Test-Parameters
{
    function Get-MissingRequiredParameters {
        $listOfMissingParameters = New-Object -TypeName 'System.Collections.ArrayList';

        #   TO DO: Enumerate below the parameters that are required
        if (ParameterIsMissing($ApiUrl)) { $listOfMissingParameters.Add("ApiUrl") | Out-Null; }
        if (ParameterIsMissing($AuthenticationUrl)) { $listOfMissingParameters.Add("AuthenticationUrl") | Out-Null; }
        if (ParameterIsMissing($OAuthClientID)) { $listOfMissingParameters.Add("OAuthClientID") | Out-Null; }
        if (ParameterIsMissing($OAuthSecret)) { $listOfMissingParameters.Add("OAuthSecret") | Out-Null; }
        if (ParameterIsMissing($ProviderName)) { $listOfMissingParameters.Add("ProviderName") | Out-Null; }
        #if (ParameterIsMissing($ProviderOwner)) { $listOfMissingParameters.Add("ProviderOwner") | Out-Null; }
        #if (ParameterIsMissing($ProviderDescription)) { $listOfMissingParameters.Add("ProviderDescription") | Out-Null; }
        if (ParameterIsMissing($EntityName)) { $listOfMissingParameters.Add("EntityName") | Out-Null; }
        #if (ParameterIsMissing($EntityRowDelimiter)) { $listOfMissingParameters.Add("EntityRowDelimiter") | Out-Null; }
        if (ParameterIsMissing($EntityFieldDelimiter)) { $listOfMissingParameters.Add("EntityFieldDelimiter") | Out-Null; }
        if (ParameterIsMissing($EntityRegularExpression)) { $listOfMissingParameters.Add("EntityRegularExpression") | Out-Null; }
        if (ParameterIsMissing($EntityStartValidDate)) { $listOfMissingParameters.Add("EntityStartValidDate") | Out-Null; }
        #if (ParameterIsMissing($EntityDescription)) { $listOfMissingParameters.Add("EntityDescription") | Out-Null; }
        if (ParameterIsMissing($AttributesCollection)) { $listOfMissingParameters.Add("AttributesCollection") | Out-Null; }
        
        return $listOfMissingParameters;
    }
    function Read-Psd1FileToObtainNeededParameters {
        if ([string]::IsNullOrWhiteSpace($ParamsFile))
        {
            Write-Warning "The ___ParamsFile___ parameter is empty, while some required parameters are missing from the command-line...";
            return;
        }
        if (-not(Test-Path -Path $ParamsFile -PathType Leaf))
        {
            Write-Warning ("The referenced ___ParamsFile___ '" + $ParamsFile + "' could not be found, while some required parameters are missing from the command-line...");
            return;
        }
        Write-Verbose ("Reading command parameters from: '" + $ParamsFile + "'");
        Get-Content $ParamsFile | Write-Verbose;
        $paramsFromFile = Import-PowerShellDataFile -Path $ParamsFile;
        
        #   TO DO: Enumerate below the parameters that could be read from the PSD1/params file
        if (ParameterIsMissing($script:ApiUrl)) { $script:ApiUrl = $paramsFromFile.SidraCore.ApiUrl; }
        if (ParameterIsMissing($script:AuthenticationUrl)) { $script:AuthenticationUrl = $paramsFromFile.SidraCore.AuthenticationUrl; }
        if (ParameterIsMissing($script:OAuthClientID)) { $script:OAuthClientID = $paramsFromFile.SidraCore.OAuthClientID; }
        if (ParameterIsMissing($script:OAuthSecret)) { $script:OAuthSecret = $paramsFromFile.SidraCore.OAuthSecret; }
        if (ParameterIsMissing($script:ProviderName)) { $script:ProviderName = $paramsFromFile.Metadata.Provider.Name; }
        if (ParameterIsMissing($script:ProviderOwner)) { $script:ProviderOwner = $paramsFromFile.Metadata.Provider.Owner; }
        if (ParameterIsMissing($script:ProviderDescription)) { $script:ProviderDescription = $paramsFromFile.Metadata.Provider.Description; }
        if (ParameterIsMissing($script:EntityName)) { $script:EntityName = $paramsFromFile.Metadata.Entity.Name; }
        if (ParameterIsMissing($script:EntityRowDelimiter)) { $script:EntityRowDelimiter = $paramsFromFile.Metadata.Entity.RowDelimiter; }
        if (ParameterIsMissing($script:EntityFieldDelimiter)) { $script:EntityFieldDelimiter = $paramsFromFile.Metadata.Entity.FieldDelimiter; }
        if (ParameterIsMissing($script:EntityRegularExpression)) { $script:EntityRegularExpression = $paramsFromFile.Metadata.Entity.RegularExpression; }
        if (ParameterIsMissing($script:EntityStartValidDate)) { $script:EntityStartValidDate = $paramsFromFile.Metadata.Entity.StartValidDate; }
        if (ParameterIsMissing($script:EntityDescription)) { $script:EntityDescription = $paramsFromFile.Metadata.Entity.Description; }
        if (ParameterIsMissing($script:AttributesCollection)) { $script:AttributesCollection = $paramsFromFile.Metadata.Attributes; }
    }
    $missingParameters = Get-MissingRequiredParameters;
    if ($missingParameters.Count -gt 0)
    {
        Write-Verbose ("Required parameters are missing from the command-line; these will be tentatively read from a ___ParamsFile___:");
        foreach ($paramName in $missingParameters) { Write-Verbose ("`t > " + $paramName); }
    }
    Read-Psd1FileToObtainNeededParameters;      # Read parameters anyway; not only if some of the required ones are missing
    $missingParameters = Get-MissingRequiredParameters;
    if ($missingParameters.Count -gt 0)
    {
        Write-Warning "The script is missing one or more required parameter(s), from both command-line and the ___ParamsFile___:";
        foreach ($paramName in $missingParameters) { Write-Warning ("`t > " + $paramName); }
        Write-Warning "This script cannot continue without needed parameters.";
        Write-Warning "The script is now prematurely exitting!...";
        return $false;
    }
    else
    {
        return $true;
    }
}
if (-not (Test-Parameters)) { return; }

#endregion  VALIDATE PARAMETERS; TENTATIVELY READ .PSD1 FILE FOR REQUIRED PARAMS THAT ARE MISSING IN COMMAND-LINE
#----------------------------------------------------------------------------------------------------------------------

# TO DO: Place additional parameter validation here; the above is only checking if required parameters are present or not.

#----------------------------------------------------------------------------------------------------------------------
#   SCRIPT CODE
#----------------------------------------------------------------------------------------------------------------------

#----------------------------------------------------------------------------------------------------------------------
#region     GET THE AUTHORIZATION TOKEN FROM THE IDENTITY SERVER
#----------------------------------------------------------------------------------------------------------------------

$HttpRequestBody = @{
    "grant_type" = "client_credentials"
    "scope" = "IdentityServerApi sidra.api"
    "client_id" = $OAuthClientID
    "client_secret" = $OAuthSecret
};
$HttpRequestHeaders = New-Object "System.Collections.Generic.Dictionary[[String],[String]]";
$HttpRequestHeaders.Add("Accept","application/json");
$HttpRequestHeaders.Add("Content-Type","application/x-www-form-urlencoded");
$IdentityServerTokenEndpoint = $AuthenticationUrl.trim("/") + "/connect/token";
Write-Verbose "Calling the Identity Server authorization token endpoint:";
Write-Verbose "Invoke-RestMethod -Uri $IdentityServerTokenEndpoint -Headers <HttpRequestHeaders> -Method 'POST' -Body <HttpRequestBody>";
Write-Verbose ($HttpRequestHeaders | ConvertTo-Json);
Write-Verbose "HTTP request payload:";
Write-Verbose ($HttpRequestBody | ConvertTo-Json);
$HttpResponseOAuth = Invoke-RestMethod -Uri $IdentityServerTokenEndpoint -Headers $HttpRequestHeaders -Method "POST" -Body $HttpRequestBody;
$OAuthToken = $HttpResponseOAuth.access_token;
if ($OAuthToken)
{
    Write-Verbose "OAuth token for calling the Sidra API:";
    Write-Verbose $OAuthToken;
}
else 
{
    Write-Warning "No OAuth token was retrieved from the Identity Server endpoint:";
    Write-Warning "`t > `t $IdentityServerTokenEndpoint";
    Write-Warning "This script cannot continue without the OAuth token.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}

#endregion  GET THE AUTHORIZATION TOKEN FROM THE IDENTITY SERVER
#----------------------------------------------------------------------------------------------------------------------

#----------------------------------------------------------------------------------------------------------------------
#region     PREPARE THE NEXT/API CALLS, BY ADDING THE AUTHORIZATION BEARER TOKEN FROM IDENTITY SERVER
#----------------------------------------------------------------------------------------------------------------------

$HttpRequestHeaders = New-Object "System.Collections.Generic.Dictionary[[String],[String]]";
$HttpRequestHeaders.Add("Authorization", "Bearer " + $OAuthToken);
$HttpRequestHeaders.Add("Accept","*/*");
$HttpRequestHeaders.Add("Content-Type","application/json");
$HttpInvokeContentType = "application/json; charset=utf-8";
Write-Verbose "All the Sidra API calls, performed from this point on, will bear the following <HttpRequestHeaders>:";
Write-Verbose ($HttpRequestHeaders | ConvertTo-Json);

#endregion  PREPARE THE NEXT/API CALLS, BY ADDING THE AUTHORIZATION BEARER TOKEN FROM IDENTITY SERVER
#----------------------------------------------------------------------------------------------------------------------

#----------------------------------------------------------------------------------------------------------------------
#region     GET THE DSU IDENTIFIER
#----------------------------------------------------------------------------------------------------------------------

$ApiEndpoint = $ApiUrl.Trim("/") + "/api/datastorageunit/DataStorageUnits";
Write-Verbose "Getting the DSU identifier:...";
Write-Verbose "Invoke-RestMethod -Uri $ApiEndpoint -Headers <HttpRequestHeaders> -Method 'GET'";
$HttpResponseDsu = Invoke-RestMethod -Uri $ApiEndpoint -Headers $HttpRequestHeaders -Method "GET";
if (-not $HttpResponseDsu.items)
{
    Write-Warning "There were no Data Storage Unit (DSU) entries returned by the call to the API endpoint:";
    Write-Host "`t > `t $ApiEndpoint";
    Write-Warning "This script cannot continue without a DSU.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}
$TargetDSU = $null;
if ($HttpResponseDsu.items.Length -lt 1)
{
    Write-Warning "There were no Data Storage Unit (DSU) entries returned by the call to the API endpoint:";
    Write-Host "`t > `t" $ApiEndpoint;
    Write-Warning "This script cannot continue without a DSU.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}
elseif ($HttpResponseDsu.items.Length -eq 1)
{
    $TargetDSU = $HttpResponseDsu.items[0];
}
else
{
    Write-Host ("There are multiple Data Storage Units (DSU) configured.") -ForegroundColor DarkYellow;
    foreach ($Entry in $HttpResponseDsu.items)
    {
        Write-Host ("$($Entry.id):`t$($Entry.name) `t $($Entry.ResourceGroupName)");
    }
    $DsuIdInput = Read-Host -Prompt "Which DSU are you targeting? >";
    $TargetDSU = $HttpResponseDsu.items | Where-Object -Property "id" -EQ -Value $DsuIdInput;
}
if ($TargetDSU)
{
    Write-Host ("Selected DSU, Data Storage Unit: `t $($TargetDSU.ID) `t>`t $($TargetDSU.Name)") -ForegroundColor Green;
}
else
{
    Write-Warning "There was no Data Storage Unit (DSU) entry selected.";
    Write-Warning "This script cannot continue without a DSU.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}

#endregion  GET THE DSU IDENTIFIER
#----------------------------------------------------------------------------------------------------------------------

$RegExNonAlphaNumeric = '[^a-zA-Z0-9]';
$ProviderDatabaseName = $ProviderName.ToLower() -Replace $RegExNonAlphaNumeric, '';
$EntityTableName = $EntityName.ToLower() -Replace $RegExNonAlphaNumeric, '_';

#----------------------------------------------------------------------------------------------------------------------
#region     CREATE THE PROVIDER OR SELECT IT IF ALREADY EXISTS
#----------------------------------------------------------------------------------------------------------------------

Write-Host "Selecting Provider '$ProviderName' if it already exists; otherwise, creating it..."
$ApiEndpoint = $ApiUrl.Trim("/") + "/api/metadata/providers";
$ProviderID = $null;
$CallUrl = $ApiEndpoint + "?field=ProviderName&text=$ProviderName&exactMatch=true";
Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'GET'";
$HttpResponseProvider = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "GET";
if ($HttpResponseProvider -and ($null -ne $HttpResponseProvider.items) -and ($HttpResponseProvider.items.Length -gt 0))
{
    $ProviderID = $HttpResponseProvider.items[0].id;
    Write-Host ("Provider '" + $HttpResponseProvider.items[0].providerName + "' already exists (ProviderID $ProviderID).") -ForegroundColor DarkGreen;
    $shouldUpdateRecord = $false;
    $shouldUpdateRecord = $shouldUpdateRecord -or ($ProviderOwner -notlike $HttpResponseProvider.items[0].owner.name);
    $shouldUpdateRecord = $shouldUpdateRecord -or ($ProviderDescription -notlike $HttpResponseProvider.items[0].description);
    if ($shouldUpdateRecord)
    {
        Write-Host "Updating the Provider's owner and description... " -ForegroundColor DarkGray -NoNewline;   
        $HttpRequestBody = @{
            "owner" = $ProviderOwner
            "description" = $ProviderDescription
        };
        $HttpRequestJson = ConvertTo-Json -InputObject $HttpRequestBody;
        $CallUrl = $ApiEndpoint + "/" + $ProviderID;
        Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'PATCH' -Body <HttpRequestJson> -ContentType $HttpInvokeContentType";
        Write-Verbose "HTTP request payload:";
        Write-Verbose ($HttpRequestBody | ConvertTo-Json);
        $HttpResponseProvider = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "PATCH" -Body $HttpRequestJson -ContentType $HttpInvokeContentType;
        Write-Host "DONE" -ForegroundColor DarkGreen;
    }
}
else
{
    $HttpRequestBody = @{
        "providerName" = $ProviderName
        "databaseName" = $ProviderDatabaseName
        "owner" = $ProviderOwner
        "description" = $ProviderDescription
        "idDataStorageUnit" = $TargetDSU.ID
    };
    $HttpRequestJson = ConvertTo-Json -InputObject $HttpRequestBody;
    Write-Verbose "Invoke-RestMethod -Uri $ApiEndpoint -Headers <HttpRequestHeaders> -Method 'POST' -Body <HttpRequestJson> -ContentType $HttpInvokeContentType";
    Write-Verbose "HTTP request payload:";
    Write-Verbose ($HttpRequestBody | ConvertTo-Json);
    $HttpResponseProvider = Invoke-RestMethod -Uri $ApiEndpoint -Headers $HttpRequestHeaders -Method "POST" -Body $HttpRequestJson -ContentType $HttpInvokeContentType;
    $ProviderID = $HttpResponseProvider.id;
    Write-Host ("Provider '" + $HttpResponseProvider.providerName + "' was created (ProviderID $ProviderID).") -ForegroundColor Green;
}
Write-Verbose ($HttpResponseProvider | Format-List | Out-String);
if (-not $ProviderID)
{
    Write-Warning "The Provider '$ProviderName' does not exist, and it could not be created either!";
    Write-Warning "This script cannot continue without a Provider ID.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}

#endregion  CREATE THE PROVIDER OR SELECT IT IF ALREADY EXISTS
#----------------------------------------------------------------------------------------------------------------------

#----------------------------------------------------------------------------------------------------------------------
#region     CREATE OR UPDATE THE ENTITY ENTRY
#----------------------------------------------------------------------------------------------------------------------

Write-Host "Creating or updating the Entity entry '$EntityName' for the CSV file(s) to be ingested...";
$ApiEndpoint = $ApiUrl.Trim("/") + "/api/metadata/entities";
$CallUrl = $ApiEndpoint + "?field=Name&text=$EntityName&exactMatch=true";
Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'GET'";
$HttpResponseEntity = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "GET";
$EntityID = $null;
$ExistingEntity = $HttpResponseEntity.items | Where-Object -Property "idProvider" -EQ -Value $ProviderID;
if ($ExistingEntity)
{
    $EntityID = $ExistingEntity.id;
    $CallUrl = $ApiEndpoint + "/" + $EntityID;
    Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'GET'";
    $HttpResponseEntity = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "GET";
    $ExistingEntity = $HttpResponseEntity;
    Write-Host "An Entity '$($ExistingEntity.Name)' with ID $EntityID exists already for the provider '$ProviderName' with ID $ProviderID." -ForegroundColor DarkGreen;
}
$EntityProperties = @"
{
    "sourceTableName": "[<Schema>].[$EntityName]",
    "consolidationMode": "Merge",
    "readerOptions": {
        "delimiter": "$EntityFieldDelimiter", 
        "encoding": "$EntityEncoding", 
        "header": true, 
        "multiLine": true
    }
}
"@;
$HttpRequestBody = @{
    "filesPerDrop" = 1
    "rowDelimiter" = $EntityRowDelimiter
    "reCreateTableOnDeployment" = $true
    "nullText" = $null
    "format" = $EntityFormat
    "lastDeployed" = $null
    "lastUpdated" = (Get-Date -Format "yyyy-MM-ddTHH:mm:ss.fffZ")
    "fieldDelimiter" = $EntityFieldDelimiter
    "headerLines" = $EntityHeaderLinesCount
    "encoding" = $EntityEncoding
    "serdeProperties" = $null
    "serde" = $null
    "endValidDate" = $null
    "startValidDate" = $EntityStartValidDate
    "regularExpression" = $EntityRegularExpression
    "tableName" = $EntityTableName
    "name" = $EntityName
    "description" = $EntityDescription
    "idProvider" = $ProviderID
    "additionalProperties" = $EntityProperties
    "generateDeltaTable" = $false
    "idTableFormat" = 3     # 1=Orc; 2=Parquet; 3=Delta
    "sourcePath" = $null
};
if ($ExistingEntity)
{
    $shouldUpdateRecord = $false;
    $shouldUpdateRecord = $shouldUpdateRecord -or ($EntityDescription -notlike $ExistingEntity.description);
    $shouldUpdateRecord = $shouldUpdateRecord -or ($EntityFieldDelimiter -notlike $ExistingEntity.fieldDelimiter);
    $shouldUpdateRecord = $shouldUpdateRecord -or ($EntityRegularExpression -ne $ExistingEntity.regularExpression);
    $shouldUpdateRecord = $shouldUpdateRecord -or ($EntityHeaderLinesCount -ne $ExistingEntity.headerLines);
    if ($shouldUpdateRecord)
    {
        Write-Host "Updating the Entity's fields... " -ForegroundColor DarkGray -NoNewline;
        $HttpRequestJson = ConvertTo-Json -InputObject $HttpRequestBody;
        $CallUrl = $ApiEndpoint + "/" + $EntityID;
        Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'PUT' -Body <HttpRequestJson> -ContentType $HttpInvokeContentType";
        Write-Verbose "HTTP request payload:";
        Write-Verbose ($HttpRequestBody | ConvertTo-Json);
        $HttpResponseEntity = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "PUT" -Body $HttpRequestJson -ContentType $HttpInvokeContentType;
        Write-Host "DONE" -ForegroundColor DarkGreen;
    }
}
else
{
    Write-Verbose "Creating the Entity entry in metadata [Core] DB:";
    $HttpRequestJson = ConvertTo-Json -InputObject $HttpRequestBody;
    Write-Verbose "Invoke-RestMethod -Uri $ApiEndpoint -Headers <HttpRequestHeaders> -Method 'POST' -Body <HttpRequestJson> -ContentType $HttpInvokeContentType";
    Write-Verbose "HTTP request payload:";
    Write-Verbose ($HttpRequestBody | ConvertTo-Json);
    $HttpResponseEntity = Invoke-RestMethod -Uri $ApiEndpoint -Headers $HttpRequestHeaders -Method "POST" -Body $HttpRequestJson -ContentType $HttpInvokeContentType;
    $EntityID = $HttpResponseEntity.id;
    Write-Host ("Entity '" + $HttpResponseEntity.name + "' was created, with Entity ID=" + $EntityID) -ForegroundColor Green;
}
if (-not $EntityID)
{
    Write-Warning "The Entity '$EntityName' (for the Provider '$ProviderName') does not exist, and it could not be created either!";
    Write-Warning "This script cannot continue without a Provider ID.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}
Write-Verbose ($HttpResponseEntity | Format-List | Out-String);

#endregion  CREATE OR UPDATE THE ENTITY ENTRY
#----------------------------------------------------------------------------------------------------------------------

#----------------------------------------------------------------------------------------------------------------------
#region     PREPARE TO CREATE OR UPDATE ATTRIBUTES
#----------------------------------------------------------------------------------------------------------------------

$ApiEndpoint = $ApiUrl.Trim("/") + "/api/metadata/attributes";
$CallUrl = $ApiEndpoint + "?field=IdEntity&text=$EntityID&exactMatch=true";
Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'GET'";
$HttpResponseAttributes = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "GET";
$ExistingAttributes = $null;
if ($HttpResponseAttributes -and ($null -ne $HttpResponseAttributes.items) -and ($HttpResponseAttributes.items.Length -gt 0))
{
    $ExistingAttributes = $HttpResponseAttributes.items;
    Write-Host "The Entity '" + $ExistingEntity.name + "' (EntityID $EntityID) from provider '$ProviderName' (ProviderID $ProviderID) already has Attributes defined." -ForegroundColor Yellow;
    Write-Host "The Attributes set you're trying to add would be merged with the existing Attributes set of the Entity."
    $choice = Read-Host -Prompt "Would you want to rather re-create all Attributes from scratch (y/N)?";
    if ($choice.StartsWith("Y", "CurrentCultureIgnoreCase"))
    {
        Write-Host ("Deleting existing Attributes for the '$EntityName' Entity...") -ForegroundColor DarkMagenta;
        $CountOfFailedDeletes = 0;
        foreach ($AttributeEntry in $ExistingAttributes)
        {
            Write-Host "Deleting Attribute ID $($AttributeEntry.ID): $($AttributeEntry.Name) ($($AttributeEntry.HiveType))..." -ForegroundColor DarkMagenta;
            $CallUrl = $ApiEndpoint + "/" + $AttributeEntry.id;
            Write-Verbose "Invoke-RestMethod $CallUrl -Headers <HttpRequestHeaders> -Method 'DELETE'";
            try
            {
                Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "DELETE";
                Write-Host "    ... deleted." -ForegroundColor DarkMagenta;
            }
            catch
            {
                Write-Host "    ... delete failed!" -ForegroundColor Yellow;
                $CountOfFailedDeletes++;
            }
        }
        if ($CountOfFailedDeletes -gt 0)
        {
            Write-Warning "Existing Attribute(s) for the Entity could not be deleted for cleanup before re-creating. The script may fail later at Attribute definition stage.";
        }
        $CallUrl = $ApiEndpoint + "?field=IdEntity&text=$EntityID&exactMatch=true";
        Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'GET'";
        $HttpResponseAttributes = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "GET";
        $ExistingAttributes = $null;
        if ($HttpResponseAttributes -and ($null -ne $HttpResponseAttributes.items) -and ($HttpResponseAttributes.items.Length -gt 0))
        {
            $ExistingAttributes = $HttpResponseAttributes.items;
        }
    }
    else
    {
        Write-Host "The Entity's Attributes will be tentatively updated, merging the existing set with the one being added... ";
    }
}
$Attribute = @{
    "sqlType" = $null
    "isMetadata" = $false
    "isPartitionColumn" = $false
    "isCalculated" = $false
    "order" = 0
    "isPrimaryKey" = $false
    "treatEmptyAsNull" = $true
    "specialFormat" = $null
    "replacementText" = $null
    "replacedText" = $null
    "removeQuotes" = $true
    "needTrim" = $true
    "isNullable" = $true
    "maxLen" = $null
    "hiveType" = $null
    "name" = $null
    "idEntity" = $EntityID
    "validationText" = $null
    "description" = $null
    "sourceName" = $null
    "isEncrypted" = $false
    #"itemId" = $null
    "idProvider" = $ProviderID
    "idDataStorageUnit" = $TargetDSU.ID
    "dataMask" = $null
    "parentSecurityPath" = "$($TargetDSU.ID)/$ProviderID/$EntityID"
    "securityPath" = $null
    "sourceType" = $null
};

function UpsertAttribute {
    param(
            [Parameter()] [hashtable] $NewAttribute,
            [Parameter()] [PSCustomObject] $OldAttribute = $null
        );
    $ApiEndpoint = $ApiUrl.Trim("/") + "/api/metadata/attributes";
    $CallUrl = $null;
    $HttpRequestJson = $null;
    $HttpResponseAttribute = $null;
    if ($OldAttribute)
    {
        $OldAttribute.sqlType =              $NewAttribute["sqlType"];
        $OldAttribute.isMetadata =           $NewAttribute["isMetadata"];
        $OldAttribute.isPartitionColumn =    $NewAttribute["isPartitionColumn"];
        $OldAttribute.isCalculated =         $NewAttribute["isCalculated"];
        $OldAttribute.order =                $NewAttribute["order"];
        $OldAttribute.isPrimaryKey =         $NewAttribute["isPrimaryKey"];
        $OldAttribute.treatEmptyAsNull =     $NewAttribute["treatEmptyAsNull"];
        $OldAttribute.specialFormat =        $NewAttribute["specialFormat"];
        $OldAttribute.replacementText =      $NewAttribute["replacementText"];
        $OldAttribute.replacedText =         $NewAttribute["replacedText"];
        $OldAttribute.removeQuotes =         $NewAttribute["removeQuotes"];
        $OldAttribute.needTrim =             $NewAttribute["needTrim"];
        $OldAttribute.isNullable =           $NewAttribute["isNullable"];
        $OldAttribute.maxLen =               $NewAttribute["maxLen"];
        $OldAttribute.hiveType =             $NewAttribute["hiveType"];
        #$OldAttribute.name =                $NewAttribute["name"];
        #$OldAttribute.idEntity =            $NewAttribute["idEntity"];
        $OldAttribute.validationText =       $NewAttribute["validationText"];
        $OldAttribute.description =          $NewAttribute["description"];
        #$OldAttribute.sourceName =          $NewAttribute["sourceName"];
        $OldAttribute.isEncrypted =          $NewAttribute["isEncrypted"];
        #$OldAttribute.itemId =              $NewAttribute["itemId"];
        #$OldAttribute.idProvider =          $NewAttribute["idProvider"];
        #$OldAttribute.idDataStorageUnit =   $NewAttribute["idDataStorageUnit"];
        $OldAttribute.dataMask =             $NewAttribute["dataMask"];
        #$OldAttribute.parentSecurityPath =  $NewAttribute["parentSecurityPath"];
        #$OldAttribute.securityPath =        $NewAttribute["securityPath"];
        $OldAttribute.sourceType =           $NewAttribute["sourceType"];
        if (($null -eq $OldAttribute.tags) -or ($OldAttribute.tags.Length -eq 0)) { $OldAttribute.tags = @(); }    #    If $null (as it comes from Core API), we'll get a '500' response
        $HttpRequestJson = ConvertTo-Json -InputObject $OldAttribute;
        $CallUrl = $ApiEndpoint + "/" + $OldAttribute.id;
        Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'PUT' -Body <HttpRequestJson> -ContentType $HttpInvokeContentType";
        $HttpResponseAttribute = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "PUT" -Body $HttpRequestJson -ContentType $HttpInvokeContentType;
    }
    else
    {
        $HttpRequestJson = ConvertTo-Json -InputObject $NewAttribute;
        $CallUrl = $ApiEndpoint;
        Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'POST' -Body <HttpRequestJson> -ContentType $HttpInvokeContentType";
        $HttpResponseAttribute = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "POST" -Body $HttpRequestJson -ContentType $HttpInvokeContentType;
    }
    return $HttpResponseAttribute;
}

#endregion  PREPARE TO CREATE OR UPDATE ATTRIBUTES
#----------------------------------------------------------------------------------------------------------------------

#----------------------------------------------------------------------------------------------------------------------
#region     CREATE OR UPDATE THE REQUIRED METADATA ATTRIBUTES, NEEDED IN DATABRICKS
#----------------------------------------------------------------------------------------------------------------------

$Attribute["isNullable"] = $false;
$Attribute["needTrim"] = $false;
$Attribute["removeQuotes"] = $false;
$Attribute["treatEmptyAsNull"] = $false;
$Attribute["isCalculated"] = $true;
$Attribute["isPartitionColumn"] = $false;
$Attribute["isMetadata"] = $true;

Write-Host "Creating or updating metadata system Attributes for Databricks:...";

$Attribute["name"] = "SidraIsDeleted";
$Attribute["hiveType"] = "BOOLEAN";
$Attribute["isNullable"] = $false;
$Attribute["specialFormat"] = "SidraIsDeleted";
$Attribute["sqlType"] = "BIT";
$Attribute["order"] = 32763;
$Attribute["isPartitionColumn"] = $false;
$ExistingAttr = $ExistingAttributes | Where-Object -Property "name" -EQ -Value $Attribute["name"];
$AttrResponse = UpsertAttribute -NewAttribute $Attribute -OldAttribute $ExistingAttr;
Write-Host ("  >  Attribute ID " + $AttrResponse.id + " in place for metadata field `t'" + $AttrResponse.name + "'`t of type '" + $AttrResponse.hiveType + "'.") -ForegroundColor DarkGray;

$Attribute["name"] = "SidraPassedValidation";
$Attribute["hiveType"] = "BOOLEAN";
$Attribute["isNullable"] = $true;
$Attribute["specialFormat"] = $null;
$Attribute["sqlType"] = "BIT";
$Attribute["order"] = 32764;
$Attribute["isPartitionColumn"] = $false;
$ExistingAttr = $ExistingAttributes | Where-Object -Property "name" -EQ -Value $Attribute["name"];
$AttrResponse = UpsertAttribute -NewAttribute $Attribute -OldAttribute $ExistingAttr;
Write-Host ("  >  Attribute ID " + $AttrResponse.id + " in place for metadata field `t'" + $AttrResponse.name + "'`t of type '" + $AttrResponse.hiveType + "'.") -ForegroundColor DarkGray;

$Attribute["name"] = "SidraLoadDate";
$Attribute["hiveType"] = "TIMESTAMP";
$Attribute["isNullable"] = $false;
$Attribute["specialFormat"] = "CURRENT_TIMESTAMP()";
$Attribute["sqlType"] = "DATETIME2(7)";
$Attribute["order"] = 32765;
$Attribute["isPartitionColumn"] = $false;
$ExistingAttr = $ExistingAttributes | Where-Object -Property "name" -EQ -Value $Attribute["name"];
$AttrResponse = UpsertAttribute -NewAttribute $Attribute -OldAttribute $ExistingAttr;
Write-Host ("  >  Attribute ID " + $AttrResponse.id + " in place for metadata field `t'" + $AttrResponse.name + "'`t of type '" + $AttrResponse.hiveType + "'.") -ForegroundColor DarkGray;

$Attribute["name"] = "SidraIdAsset";
$Attribute["hiveType"] = "INT";
$Attribute["isNullable"] = $false;
$Attribute["specialFormat"] = "SidraIdAsset";
$Attribute["sqlType"] = "INT";
$Attribute["order"] = 32766;
$Attribute["isPartitionColumn"] = $true;
$ExistingAttr = $ExistingAttributes | Where-Object -Property "name" -EQ -Value $Attribute["name"];
$AttrResponse = UpsertAttribute -NewAttribute $Attribute -OldAttribute $ExistingAttr;
Write-Host ("  >  Attribute ID " + $AttrResponse.id + " in place for metadata field `t'" + $AttrResponse.name + "'`t of type '" + $AttrResponse.hiveType + "'.") -ForegroundColor DarkGray;

$Attribute["name"] = "SidraFileDate";
$Attribute["hiveType"] = "DATE";
$Attribute["isNullable"] = $false;
$Attribute["specialFormat"] = "SidraFileDate";
$Attribute["sqlType"] = "DATE";
$Attribute["order"] = 32767;
$Attribute["isPartitionColumn"] = $true;
$ExistingAttr = $ExistingAttributes | Where-Object -Property "name" -EQ -Value $Attribute["name"];
$AttrResponse = UpsertAttribute -NewAttribute $Attribute -OldAttribute $ExistingAttr;
Write-Host ("  >  Attribute ID " + $AttrResponse.id + " in place for metadata field `t'" + $AttrResponse.name + "'`t of type '" + $AttrResponse.hiveType + "'.") -ForegroundColor DarkGray;

#endregion  CREATE OR UPDATE THE REQUIRED METADATA ATTRIBUTES, NEEDED IN DATABRICKS
#----------------------------------------------------------------------------------------------------------------------

#----------------------------------------------------------------------------------------------------------------------
#region     CREATE OR UPDATE THE ATTRIBUTES CORRESPONDING TO CSV RECORD FIELDS
#----------------------------------------------------------------------------------------------------------------------

$Attribute["specialFormat"] = $null;
$Attribute["isNullable"] = $true;
$Attribute["needTrim"] = $true;
$Attribute["removeQuotes"] = $true;
$Attribute["treatEmptyAsNull"] = $true;
$Attribute["isCalculated"] = $false;
$Attribute["isPartitionColumn"] = $false;
$Attribute["isMetadata"] = $false;

Write-Host "Creating the Attributes for columns found in the CSV definition:...";
$AttributeOrder = 1;
foreach ($CsvColumn in $AttributesCollection)
{
    $Attribute["order"] = $AttributeOrder++;
    $Attribute["name"] = $CsvColumn.ColumnName;
    $Attribute["hiveType"] = $CsvColumn.HiveType;
    $Attribute["sqlType"] = $CsvColumn.SqlType;
    $Attribute["isPrimaryKey"] = $CsvColumn.IsPrimaryKey;
    $ExistingAttr = $ExistingAttributes | Where-Object -Property "name" -EQ -Value $Attribute["name"];
    $AttrResponse = UpsertAttribute -NewAttribute $Attribute -OldAttribute $ExistingAttr;
    Write-Host ("  >  Attribute ID " + $AttrResponse.id + " in place for `t'" + $AttrResponse.name + "'`t of type '" + $AttrResponse.hiveType + "'.") -ForegroundColor Green;
}

#endregion  CREATE OR UPDATE THE ATTRIBUTES CORRESPONDING TO CSV RECORD FIELDS
#----------------------------------------------------------------------------------------------------------------------

#----------------------------------------------------------------------------------------------------------------------
#region     DETERMINE THE TARGET DATA FACTORY
#----------------------------------------------------------------------------------------------------------------------

Write-Host "Getting the Data Factory pertaining to DSU '$($TargetDSU.Name)'...";
$ApiEndpoint = $ApiUrl.Trim("/") + "/api/datafactory/datafactories";
$CallUrl = $ApiEndpoint + "?text=$($TargetDSU.ResourceGroupName)&field=resourceGroup&exactMatch=true";
Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'GET'";
$HttpResponseDataFactory = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "GET";
if (-not $HttpResponseDataFactory.items)
{
    Write-Warning "There were no Data Factory entries returned by the call to the API endpoint:";
    Write-Host "`t > `t" $CallUrl;
    Write-Warning "This script cannot continue determining the proper ingestion Pipeline.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}
$TargetDataFactory = $null;
if ($HttpResponseDataFactory.items.Length -lt 1)
{
    Write-Warning "There were no '$($TargetDSU.ResourceGroupName)' Data Factory entries returned by the call to the API endpoint:";
    Write-Host "`t > `t" $CallUrl;
    Write-Warning "This script cannot continue determining the proper ingestion Pipeline.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}
else
{
    $TargetDataFactory = $HttpResponseDataFactory.items[0];
}
if (-not $TargetDataFactory)
{
    Write-Warning "There was no Data Factory entry having 'ResourceGroup' equal to '$($TargetDSU.ResourceGroupName)'.";
    Write-Warning "This script cannot continue determining a proper ingestion Pipeline.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}
Write-Host "Data Factory for DSU $($TargetDSU.ID), '$($TargetDSU.Name)': $($TargetDataFactory.ID), '$($TargetDataFactory.Name)'." -ForegroundColor DarkGreen;

#endregion  DETERMINE THE TARGET DATA FACTORY
#----------------------------------------------------------------------------------------------------------------------

#----------------------------------------------------------------------------------------------------------------------
#region     ASSOCIATE THE ENTITY WITH THE INGESTION PIPELINE: FileIngestionDatabricks
#----------------------------------------------------------------------------------------------------------------------

Write-Host "Getting the ingestion Pipeline for CSV files, 'FileIngestionDatabricks'...";
$ApiEndpoint = $ApiUrl.Trim("/") + "/api/datafactory/pipelines";
$CallUrl = $ApiEndpoint + "?text=FileIngestionDatabricks&field=Name&exactMatch=true";
Write-Verbose "Invoke-RestMethod -Uri $CallUrl -Headers <HttpRequestHeaders> -Method 'GET'";
$HttpResponsePipeline = Invoke-RestMethod -Uri $CallUrl -Headers $HttpRequestHeaders -Method "GET";
if (-not $HttpResponsePipeline.items)
{
    Write-Warning "There were no Pipeline entries returned by the call to the API endpoint:";
    Write-Host "`t > `t" $CallUrl;
    Write-Warning "This script cannot continue associating the Entity with the proper ingestion Pipeline.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}
$Pipeline = $null;
if ($HttpResponsePipeline.items.Length -lt 1)
{
    Write-Warning "There were no 'FileIngestionDatabricks' Pipeline entries returned by the call to the API endpoint:";
    Write-Host "`t > `t" $CallUrl;
    Write-Warning "This script cannot continue associating the Entity with the proper ingestion Pipeline.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}
else
{
    $Pipeline = $HttpResponsePipeline.items | Where-Object -Property "idDataFactory" -EQ -Value $($TargetDataFactory.ID);
}
if (-not $Pipeline)
{
    Write-Warning "There was no 'FileIngestionDatabricks' Pipeline entry having 'idDataFactory' equal to $($TargetDataFactory.ID).";
    Write-Warning "This script cannot continue associating the Entity with a proper ingestion Pipeline.";
    Write-Warning "The script is now prematurely exitting!...";
    return;
}
Write-Host "Pipeline for Data Factory $($TargetDataFactory.ID), '$($TargetDataFactory.Name)': $($Pipeline.ID), '$($Pipeline.Name)'." -ForegroundColor DarkGreen;
Write-Host "Associating the Entity '$EntityName' (EntityID $EntityID) with the ingestion Pipeline 'FileIngestionDatabricks' (PipelineID $($Pipeline.ID), of DSU $($TargetDSU.ID):$($TargetDSU.Name))..."
Write-Host "Checking if association already exists... " -NoNewline;
$ApiEndpoint = $ApiUrl.Trim("/") + "/api/metadata/entities/$EntityID/pipelines";
$ExistingAssociationEntry = $null;
Write-Verbose "Invoke-RestMethod $ApiEndpoint -Headers <HttpRequestHeaders> -Method 'GET'";
$HttpResponseEntityPipeline = Invoke-RestMethod $ApiEndpoint -Headers $HttpRequestHeaders -Method "GET";
if ($HttpResponseEntityPipeline -and ($HttpResponseEntityPipeline.Length -gt 0))
{
    $ExistingAssociationEntry = $HttpResponseEntityPipeline | Where-Object -Property "idPipeline" -EQ -Value $Pipeline.ID;
    if ($ExistingAssociationEntry)
    {
        Write-Host "It does exist." -ForegroundColor DarkGreen;
        Write-Host "Association '$ProviderName>$EntityName' (EntityID $EntityID) <-> 'FileIngestionDatabricks' (PipelineID $($Pipeline.ID),  DSU $($TargetDSU.ID):$($TargetDSU.Name)) exists." -ForegroundColor Green;
    }
}
if (-not $ExistingAssociationEntry)
{
    Write-Host "Nope, needs created; doing that now..." -ForegroundColor DarkMagenta;
    $ApiEndpoint = $ApiUrl.Trim("/") + "/api/metadata/entities/$EntityID/pipelines/$($Pipeline.ID)";
    Write-Verbose "Invoke-RestMethod $ApiEndpoint -Headers <HttpRequestHeaders> -Method 'POST'";
    $HttpResponseEntityPipeline = Invoke-RestMethod $ApiEndpoint -Headers $HttpRequestHeaders -Method "POST";
    if (-not $HttpResponseEntityPipeline)
    {
        Write-Warning "No Entity <-> Pipeline association entry could be created by the call to the API endpoint:";
        Write-Host "`t > `t" $ApiEndpoint;
        Write-Warning "Metadata was created for the '$ProviderName>$EntityName' Entity, but no Pipeline association has been made yet. This may need to be added manually.";
        Write-Warning "The script is now prematurely exitting!...";
        return;
    }
    Write-Host "Entity '$ProviderName>$EntityName' (EntityID $EntityID) is now associated with ingestion Pipeline 'FileIngestionDatabricks' (PipelineID $($Pipeline.ID), of DSU $($TargetDSU.ID):$($TargetDSU.Name))..." -ForegroundColor Green;
}
Write-Host "Everything is done now. The Entity metadata is in place, so the ingestion pipeline ""knows"" how to transfer data into Databricks.";

#endregion  ASSOCIATE THE ENTITY WITH THE INGESTION PIPELINE: FileIngestionDatabricks
#----------------------------------------------------------------------------------------------------------------------

Write-Host "You may start a test-drive - place a CSV file with the defined structure into the Staging area:" -ForegroundColor DarkGray;
Write-Host "DSU '$($TargetDSU.Name)' > Azure Storage Account for Staging > Container " -NoNewline -ForegroundColor DarkGray;
Write-Host "landing" -NoNewline -ForegroundColor DarkCyan;
Write-Host " > Folder " -NoNewline -ForegroundColor DarkGray;
Write-Host "$ProviderName/$EntityName" -ForegroundColor DarkCyan;
